import numpy as np
from scipy.sparse import csc_matrix
from scipy.sparse.linalg import splu
from numpy.linalg import qr
from ..interface.BaseInterface import BaseInterface as bs
from .Iteral import Iteral as Int
import scipy.sparse as sp

_DEFAULT_OPTS = {
        'stop': 'matrix',
        'maxit': 1e4,
        'tol': 1e-8,
        'which': 'SM',
        'verbose': True,
    }


    
def _Z(X, A = None, which = 'SM'):
    if sp.issparse(A) == False:
        A = csc_matrix(A)
    if which == 'SM':
        z = splu(A).solve(X)
    else:
        z = A @ X
    return z

@bs.set_defaults(default_cls = Int, default_opts=_DEFAULT_OPTS)
def osim(A, V = None, k = None, **kwargs):
    """
    Orthogonal Subspace Iteration Method (OSIM).

    Args:
    A : numpy.ndarray
        The matrix to compute the eigenvalues and eigenvectors for.
    V : numpy.ndarray
        Initial guess of eigenvectors. Default None,
    k : int
        number of desired eigenvalues. Default None. 
    

    Keyword Arguments:
        tol : float, optional
            Tolerance for convergence (default: 1e-8).
        maxit : int, optional
            Maximum number of iterations (default: 100).
        stop : str, optional
            Stopping criteria for convergence. Options are 'eig' (default), 'matrix', or 'residual'.
        which : str, optional
            If small or biggest eigenvalues. Default "SM" = smallest. 
        verbose : bool, optional
            If True, prints iteration information (default: True).

    Returns:
    numpy.ndarray
        Matrix of eigenvectors.
    numpy.ndarray
        Matrix of transformed eigenvectors.
    tuple
        Tuple containing eigenvalues at each iteration.

    Notes:
    This function implements the Orthogonal Subspace Iteration Method (OSIM) to compute eigenvectors and eigenvalues of a matrix A.

    Examples:
    >>> import numpy as np
    >>> from byma.interal import osim
    >>> A = np.array([[1, 0], [0, 1]])
    >>> V = np.array([[1], [0]])
    >>> V, BV, iter = osim(A, V, tol=1e-8, maxit=1000, stop='eig')

    You can also pass keyword arguments using a dictionary. For example:
    >>> kwargs = {'tol': 1e-8, 'maxit': 1000, 'stop': 'eig'}
    >>> V, BV, iter = osim(A, V, **kwargs)

    You can also pass keyword arguments using two separate dictionaries for parameters and interface. For example:
    >>> parameters = {'tol': 1e-8, 'maxit': 1000, 'stop': 'eig'}
    >>> interface = {'verbose': True}
    >>> V, BV, iter = sosim(A, V, parameters=parameters, interface=interface)
    """
    
    _opts = bs.opts(**kwargs)
    verbose = _opts['verbose']
    tol = _opts['tol']
    maxit = int(_opts['maxit'])
    stop = _opts['stop']
    which = _opts['which']
    
    # Check if the number of rows of V matches the number of columns of A
    if V != None:
        if V.shape[0] != A.shape[1]:
            raise ValueError("Number of rows of V must match the number of columns of A.")
    else:
        if k == None:
            raise ValueError("Either V or k should be not-None")
        else:
            V = np.random.rand(A.shape[1], k)
    
    if verbose:
        print('------ OSIM initialization summary ------')
        print(f'tollerence: {tol}')
        print(f'maximum iter: {maxit}')
        print(f'stopping criteria: {stop}')
        print(f'Required eig: {which}')
    
        print('------ Start iteration ------')

    B = lambda V: V.T @ _Z(X = V, A = A, which= which)

    iter = []
    for n in range(maxit):
        Zn = _Z(X = V, A = A, which= which)
        BV = B(V = V)
        eig = np.diag(BV)
        if which == 'SM':
            eig = 1/eig
        if verbose != False: 
            if (n % verbose == 0):
                print(f"Eigenvalues at n = {n}: {eig}")
        else:
            if (n % 5000 == 0):
                print(f"iteration n = {n}")
        
        if n > 0:
            iter.append(eig)
        
        V, _ = qr(Zn)

        if stop == 'eig':
            if n > 1:
                if abs(np.linalg.norm(eig - iter[n - 1])) < tol:
                    print(f'The Orthogonal Subspace Method has converged in {n} iterations.')
                    break
            else:
                pass
        elif stop == 'matrix':
            
            if n > 1:
                if np.linalg.norm(BV - B(V)) < tol:
                    print(f'The Orthogonal Subspace Method has converged in {n} iterations.')
                    break
        
        elif stop == 'residual':
            
            if np.linalg.norm(A @ V - V @ B(V)) < tol:
                print(f'The Orthogonal Subspace Method has converged in {n} iterations.')
                break
    
        if n >= maxit:
            print('The Orthogonal Subspace Method has not converged')
        
    return V, BV, tuple(iter)
        
        

    
    


"""
Time Module
===========

This module provides functionality for measuring and recording the execution time of functions or code blocks. It offers methods to annotate functions for time measurement, record time statistics, save data to HDF5 files, and more.

Examples
============
    >>> from byma.interface import TimeRecorder, time, set_quiet, read_data

    >>> # Create a TimeRecorder object with quiet mode enabled
    >>> recorder = TimeRecorder(quiet_mode=True)

    >>> # Annotate a function to measure its execution time
    >>> @time
    ... def my_function():
    ...     time.sleep(1)
    ...
    >>> my_function()
    >>> # Output depends on the quiet mode setting

    >>> # Measure execution time using a context manager
    >>> with time("Block1"):
    ...     time.sleep(0.5)
    ...
    Block1 took 500.12ms

    >>> # Generate and print a report of recorded time statistics
    >>> recorder.report()
    time report...
                     Block1:  500.12ms for      1 calls
                 my_function: 1000.23ms for      1 calls

    >>> # Reset recorded time statistics
    >>> recorder.reset()

    >>> # Set quiet mode to suppress printing time measurements
    >>> set_quiet(True)
    >>> my_function()
    >>> # No output since quiet mode is enabled

    >>> # Save time measurements to an HDF5 file
    >>> with time(save_to_h5=True):
    ...     time.sleep(0.3)
    ...
    >>> # Data saved to 'timings.h5' file

    >>> # Read data from an HDF5 file
    >>> data = read_data('timings.h5', ['Block1'])
    >>> print(data)
    {'Block1': array(500.12)}
    
Acknowledgement
===================

This module has been made by taking inspiration and multiple parts of the code from the `timebudget` package.

"""

import sys
from time import time as pytime
import warnings
from collections import defaultdict
from functools import wraps
from typing import Callable, Optional, Union
import h5py

class TimeRecorder:
    def __init__(self, quiet_mode: bool = False):
        """
        Initialize a TimeRecorder object.

        Parameters:
        -----------
        quiet_mode : bool, optional
            Whether to suppress printing time measurements. Default is False.
            
        Examples:
        ---------
        >>> recorder = TimeRecorder(quiet_mode=True)
        """

        self.quiet_mode = quiet_mode
        self.reset()
        self.out_stream = sys.stdout

    def reset(self):
        """Clear all recorded time statistics."""
        self.start_times = {}
        self.elapsed_total = defaultdict(float)
        self.elapsed_cnt = defaultdict(int)

    def _print(self, msg: str):
        """Print a message."""
        self.out_stream.write(msg + "\n")
        self.out_stream.flush()

    def start(self, block_name: str):
        """
        Record the start time of a code block.

        Parameters:
        -----------
        block_name : str
            Name of the code block.
            
        Examples:
        ---------
        >>> recorder.start("my_block")
        """

        if block_name in self.start_times:
            warnings.warn(f"time is confused: time.start({block_name}) without end")
        self.start_times[block_name] = pytime()

    def end(self, block_name: str, quiet: Optional[bool] = None) -> float:
        """
        Record the end time of a code block and calculate the elapsed time.

        Parameters:
        -----------
        block_name : str
            Name of the code block.
        quiet : bool, optional
            Whether to suppress printing the time measurement. Default is None.

        Returns:
        --------
        float
            Elapsed time in milliseconds.
            
        Examples:
        ---------
        >>> elapsed_time = recorder.end("my_block")
        """

        if quiet is None:
            quiet = self.quiet_mode
        if block_name not in self.start_times:
            warnings.warn(f"time is confused: time.end({block_name}) without start")
            return float('NaN')
        elapsed = 1000 * (pytime() - self.start_times[block_name])
        self.elapsed_total[block_name] += elapsed
        self.elapsed_cnt[block_name] += 1
        del self.start_times[block_name]
        if not quiet:
            self._print(f"{block_name} took {self._format_time(elapsed)}")
        return elapsed

    def _format_time(self, milliseconds: float) -> str:
        """
        Format elapsed time.

        Parameters:
        -----------
        milliseconds : float
            Elapsed time in milliseconds.

        Returns:
        --------
        str
            Formatted elapsed time.
            
        Examples:
        ---------
        >>> formatted_time = recorder._format_time(1000.234)
        """

        return f"{milliseconds:.3f}ms" if milliseconds < 1 else f"{milliseconds:.2f}ms" if milliseconds < 1000 else f"{milliseconds/1000:.3f}sec"

    def report(self, percent_of: str = None, reset: bool = False):
        """
        Generate and print a report of recorded time statistics.

        Parameters:
        -----------
        percent_of : str, optional
            Name of the code block for which to generate the report as a percentage. Default is None.
        reset : bool, optional
            Whether to reset the recorded time statistics after generating the report. Default is False.
            
        Examples:
        ---------
        >>> recorder.report()
        """

        results = []
        for name, cnt in self.elapsed_cnt.items():
            total = self.elapsed_total[name]
            results.append({
                'name': name,
                'total': total,
                'cnt': cnt,
                'avg': total / cnt,
            })
        results = sorted(results, key=lambda r: r['total'], reverse=True)
        if percent_of:
            assert percent_of in self.elapsed_cnt, f"Can't generate report for unrecognized block {percent_of}"
            self._print(f"time report per {percent_of} cycle...")
            total_elapsed = self.elapsed_total[percent_of]
            total_cnt = self.elapsed_cnt[percent_of]
            for res in results:
                avg = res['total'] / total_cnt
                pct = 100.0 * res['total'] / total_elapsed
                avg_cnt = res['cnt'] / total_cnt
                self._print(f"{res['name']:>25s}:{pct: 6.2f}% {avg: 8.2f}ms/cyc @{avg_cnt: 8.1f} calls/cyc")
        else:
            self._print("time report...")
            for res in results:
                self._print(f"{res['name']:>25s}:{res['total']: 8.2f}ms for {res['cnt']: 6d} calls")
        if reset:
            self.reset()

_default_recorder = TimeRecorder()

def annotate(func: Callable, quiet: Optional[bool]):
    """
    Annotate a function or code block to measure its execution time.

    Parameters:
    -----------
    func : callable
        Function or code block to annotate.
    quiet : bool, optional
        Whether to suppress printing the time measurement. Default is None.
        
    Examples:
    ---------
    >>> @annotate(quiet=True)
    ... def my_function():
    ...     pass
    """

    name = func.__name__
    @wraps(func)
    def inner(*args, **kwargs):
        _default_recorder.start(name)
        try:
            return func(*args, **kwargs)
        finally:
            _default_recorder.end(name, quiet)
    return inner

class _timeblock():
    def __init__(self, name: str, quiet: Optional[bool]):
        self.name = name
        self.quiet = quiet

    def __enter__(self):
        _default_recorder.start(self.name)

    def __exit__(self, typ, val, trace):
        _default_recorder.end(self.name, self.quiet)

def time(func_or_name: Union[Callable, str] = None, quiet: Optional[bool] = None, save_to_h5: Optional[bool] = None, filename: Optional[str] = None, rewrite: Optional[bool] = None, plot_title: Optional[str] = None, h5_graph_title: Optional[str] = None, **kwargs):
    """
    Measure the execution time of a function or code block.

    Parameters:
    -----------
    func_or_name : callable or str, optional
        Function or code block to measure execution time. Default is None.
    quiet : bool, optional
        Whether to suppress printing the time measurement. Default is None.
    save_to_h5 : bool, optional
        Whether to save the time measurements to an HDF5 file. Default is None.
    h5_filename : str, optional
        Name of the HDF5 file to save time measurements. Default is None.
    h5_rewrite : bool, optional
        Whether to overwrite the existing HDF5 file. Default is None.
    h5_plot_title : str, optional
        Title of the plot generated from the HDF5 file. Default is None.
    h5_graph_title : str, optional
        Title of the graph generated from the HDF5 file. Default is None.
    kwargs : dict, optional
        Additional keyword arguments for customization.

    Returns:
    --------
    function or _timeblock
        Annotated function or context manager.
        
    Examples:
    ---------
    >>> @time
    ... def my_function():
    ...     pass
    """

    save_to_h5 = kwargs.get('save_to_h5', False) if save_to_h5 is None else save_to_h5
    filename = kwargs.get('filename', 'timings.h5') if filename is None else filename
    rewrite = kwargs.get('rewrite', True) if rewrite is None else rewrite
    plot_title = kwargs.get('plot_title', 'Timing Data') if plot_title is None else plot_title
    h5_graph_title = kwargs.get('h5_graph_title', 'Functions') if h5_graph_title is None else h5_graph_title

    if isinstance(func_or_name, str):  # If block name is provided
        return _timeblock(func_or_name, quiet)
    else:  # If a function is provided
        def wrapper(func):
            @wraps(func)
            def inner(*args, **kwargs):
                result = None
                block_name = func.__name__
                _default_recorder.start(block_name)
                try:
                    result = func(*args, **kwargs)
                    return result
                finally:
                    elapsed_time = _default_recorder.end(block_name, quiet)
                    if save_to_h5 and filename:
                        with h5py.File(filename, 'a') as f:
                            if rewrite:
                                if block_name in f:
                                    del f[block_name]
                            f.create_dataset(block_name, data=elapsed_time)

            return inner

        return wrapper(func_or_name)

def set_quiet(quiet: bool = True):
    """
    Set the quiet mode for suppressing time measurements.

    Parameters:
    -----------
    quiet : bool, optional
        Whether to suppress printing the time measurement. Default is True.
        
    Examples:
    ---------
    >>> set_quiet(True)
    """

    _default_recorder.quiet_mode = quiet

def read_data(filename: str, data_names: list):
    """
    Read data from an HDF5 file.

    Parameters:
    -----------
    filename : str
        Name of the HDF5 file.
    data_names : list
        List of data names to read from the HDF5 file.

    Returns:
    --------
    dict
        Dictionary containing the data read from the HDF5 file.
        
    Examples:
    ---------
    >>> data = read_data('timings.h5', ['Block1'])
    """

    data = {}
    with h5py.File(filename, 'r') as f:
        for name in data_names:
            if name in f:
                data[name] = f[name][()]
    return data



import numpy as np
from ..interface.BaseInterface import BaseInterface as bs
from .Iteral import Iteral as Int
from numpy import linalg as npling
from scipy.optimize import line_search
import warnings
import scipy.sparse as sp

_DEFAULT_OPTS = {
    'stop': 'normal', 
    'maxit': 1e4, 
    'tol': 1e-8, 
    'verbose': False,
    'mode': None, 
    'method': 'normal',
    'delta_x': 0.0005,
    "dim": 1,
    "domain": None, 
    "l": 1,
    "alpha": None,
    "gamma": 1,
    }


def _returns(x, fnorm, dxnorm, mode, method):
    """
    Helper function to format the return value based on the mode.

    Parameters
    ----------
    x : array_like
        Root obtained after iterations.
    fnorm : float
        Norm of the residuals.
    dxnorm : float
        Norm of the correction.
    mode : str
        Mode of the output ('full', 'partial', None).
    method : str
        Stopping method used.

    Returns
    -------
    tuple
        Tuple containing the formatted output based on the mode.
    """
    
    if mode == 'full':
        return x, dxnorm, fnorm
    elif mode == 'partial':
        if method == 'normal': 
            return x, dxnorm 
        else: 
            return x, fnorm
        
    elif ((mode == None) or (mode == False)):
        return x, dxnorm, fnorm


def calc_numerical_gradient(f, x, delta_x):
    """Function for computing gradient numerically."""
    val_at_x = f(x)
    val_at_next = f(x + delta_x)
    return (val_at_next - val_at_x) / delta_x
                

@bs.set_defaults(default_cls = Int, default_opts=_DEFAULT_OPTS)
def nesterov(x, f, df, **kwargs):
    """
    Perform Nesterov's acceleration method iterations to find the minimizer of a given function.

    Parameters
    ----------
    x : array_like
        Initial guess.
    f : callable
        Function to minimize.
    df : callable/Null
        Jacobian matrix.
    **kwargs : dict
        Additional keyword arguments for customization.
        
        domain : collable. 
            Domain. If collable, input x return Boolean. True if inside domain. Default R^n
        dim :   int. 
            Diamension starting space. Default R. 
        tol : float, optional
            Tolerance for convergence. Default is 1e-8.
        maxit : int, optional
            Maximum number of iterations. Default is 10000.
        verbose : bool, optional
            If True, prints iteration information. Default is False.
        mode : str, optional
            Mode of the output ('full', 'partial', None).
            
        stop :  int. Default 0. Default optimal 
                    0: ||grad(f)||<tol
        L : float
            smooth constant
    
    Returns
    -------
    minimize, correction_norm, residuals_norm : tuple
        If mode is 'full'
    minimize, correction_norm, (residuals_norm) : tuple
        if mode is 'partial'. The residuals_norm are returned if method is not 'normal' 
    
    minimize, iterations, correction_norm, residuals_norm: tuple
        if mode is None

    Raises
    ------
    ValueError
        If the maximum number of iterations or tolerance is not a positive integer.

    Examples
    --------
    Basic usage:
    """   

    # Initialize the known constants
    _opts = bs.opts(**kwargs)
    verbose = _opts['verbose']
    mode = _opts['mode']
    tol = _opts['tol']
    maxit = int(_opts['maxit'])
    stop = _opts['stop']
    delta_x = _opts['delta_x']
    dim = _opts['dim']
    domain = _opts['domain'] 
    L = _opts['l']
    gamma = _opts['gamma']
    lambda_prev = 0
    lambda_curr = 1
    y_prev = x
    alpha = _opts['alpha'] if _opts['alpha'] != None else 0.05 / (2 * L)
    
    
    ##### Checking correctness parameters #####
    
    assert isinstance(dim, int), "dim must be an integer"
    
    if maxit <= 0 or not isinstance(maxit, int):
        raise ValueError("Maximum number of iterations 'maxit' must be a positive integer.")
    if tol <= 0:
        raise ValueError("Tolerance 'tol' must be a positive value.")
    
    if df == None:
        assert delta_x > 0, "Step must be positive."
        df = lambda x: calc_numerical_gradient(f, x, delta_x)
        
    if verbose:
        print('------ Nesterov Method summary ------')
        print(f'tollerence: {tol}')
        print(f'maximum iter: {maxit}')
        print(f'stopping criteria: {stop}')
        print(f'starting guess: {x}')
        print(f'alpha: {alpha}')
        print(f'L: {L}')
        print(f'gamma: {gamma}')
    
        print('------ Start iteration ------')
        
    # initialize solution lists
    f_value = f(x)
    df_value = df(x)
    normdx = []
    normf = []
    
    
    for iter in range(maxit):
        
        # Initialize the prescibe type of method 
        y_curr = x - alpha * df(x)
        x = (1 - gamma) * y_curr + gamma * y_prev
        y_prev = y_curr

        lambda_tmp = lambda_curr
        lambda_curr = (1 + np.sqrt(1 + 4 * lambda_prev * lambda_prev)) / 2
        lambda_prev = lambda_tmp

        gamma = (1 - lambda_prev) / lambda_curr
    
        
        if sp.issparse(f_value) and sp.issparse(df_value):
            fnorm = sp.linalg.norm(f(x))
            dxnorm = sp.linalg.norm(df(x))
        else:
            fnorm = np.linalg.norm(f(x))
            dxnorm = np.linalg.norm(df(x))
            
        normdx.append(dxnorm)
        normf.append(fnorm)
        
        if  (verbose != 0 and verbose != False) and (iter % verbose == 0):
            print(f"Nesterov method  status at iteration {iter + 1}: ||dx|| = {dxnorm} and ||F|| = {fnorm}")
        
        if (fnorm < tol and stop == 'residual-check'):
            if verbose:
                print(f'Nesterov method converged in {iter + 1} iterations with ||F|| = {fnorm}')
            return _returns(x, fnorm=fnorm, dxnorm=normdx, mode=mode, method=stop)
        
        if (dxnorm < tol and stop == 'normal'):
            if verbose:
                print(f'Nesterov method  converged in {iter + 1} iterations with ||dx|| = {dxnorm}')
            return _returns(x, fnorm=fnorm, dxnorm=normdx, mode=mode, method=stop)
       
        if domain != None:
            if domain(x) == False:
                print('Guess is out of bounds')
                break
        
    if verbose:
        print(f'Nesterov method did not converge within {maxit} iterations')

    return _returns(x, fnorm=fnorm, dxnorm=normdx, mode=mode, method=stop)


from .Newton import *
from .OrthogonalSubspace import *
from .GradientDescent import *
from .NesterovMethod import *
from .Iteral import Iteral

__all__ = ['Newton', 'Iteral', 'OrthogonalSubspace', 'GradientDescent', 'NesterovMethod']

    